defmodule FluxRedis.PipelineManager do
  @moduledoc """
  Manage multiple commands for a single request.

  ## Command modules

  The functions of this module accepts a `t:list/1` of `t:tuple/0` with
  `t:command/0` and `t:parser/0`, both are automatically generated by the
  following modules:

  - `FluxRedis.HashManager.Commands` - Commands to manage hashes and hash
    fields.

  ## Pipeline Options

    - `:as_transaction?` - If `false`, the request will not be handled as a
      transaction. Defaults to `false`. Accepts `t:boolean/0`. For more
      information, please check
      [transactions or pipeling?](https://hexdocs.pm/redix/Redix.html#module-transactions-or-pipelining)
      section from `Redix` module.

    - `:no_reply?` - If `true`, the function returns `:ok` after successfully
      send the request to Redis. Defaults to `false`. Accepts `t:boolean/0`.

    - `:timeout` - Milliseconds to wait for a request sending. Defaults to
      `5_000`. Accepts `t:integer/0` or `:infinity`.
  """
  @moduledoc since: "0.0.2"

  alias FluxRedis.{Error, Executer}

  @typedoc "Command specs generated by one of `FluxRedis` commands modules."
  @type command :: list

  @typedoc """
  Parser function generated by one of the `FluxRedis` commands modules.
  """
  @type parser :: (redix_data -> any)

  @typedoc "Result received after a `Redix` request."
  @type redix_data :: nil | String.t() | integer | Redix.Error.t() | list(redix_data)

  @doc """
  Execute one or more commands in a single request.

  Returns `:ok` in case of successful request and `no_reply?` is set to `true`.

  Returns `{:ok, results}` in case of successful request and `:no_reply?` is
  set to `false`, where each result is defined according to the `t:parser/0`
  or `t:FluxRedis.Error.t/0` if the command failed to be processed.

  Returns `{:error, %FluxRedis.Error{}}` if request failed to reach Redis.

  ## Examples

      iex> hash_storages = [
      ...>   animals: [:phylum, :family],
      ...>   cars: [:brand, :model],
      ...>   users: :id
      ...> ]
      ...> Application.put_env(:flux_redis, :hash_storages, hash_storages)
      ...> animals = [
      ...>   %{family: "Canidae", phylum: "Chordata"},
      ...>   %{family: "Felidae", phylum: "Chordata"}
      ...> ]
      ...> car = %{brand: "Honda", model: "Civic", year: 2019}
      ...> users = [%{id: 1, name: "John"}, %{id: 2, name: "Doe"}]
      ...> pipeline_payload = [
      ...>   FluxRedis.HashManager.Commands.set_many!(:animals, animals)
      ...>   FluxRedis.HashManager.Commands.set!(:cars, car)
      ...>   FluxRedis.HashManager.Commands.set_many!(:users, users)
      ...>   FluxRedis.HashManager.Commands.delete(:users, 1)
      ...>   FluxRedis.HashManager.Commands.count(:animals)
      ...> ]
      ...> FluxRedis.PipelineManager.pipeline(pipeline_payload)
      {:ok, [2, :set, 2, :ok, 2]}
  """
  @doc since: "0.0.2"
  @spec pipeline(list({command, parser}), keyword) :: :ok | {:ok, list(any)} | {:error, Error.t()}
  def pipeline(pipeline_payload, opts \\ []) do
    pipeline_payload
    |> Enum.unzip()
    |> Executer.execute_pipeline(opts)
  end

  @doc """
  Execute one or more commands in a single request.

  Similar to `pipeline/2`, but raises `t:FluxRedis.Error.t/0` on failure (
  either from a command or from the request itself).

  ## Examples

      iex> hash_storages = [
      ...>   animals: [:phylum, :family],
      ...>   cars: [:brand, :model],
      ...>   users: :id
      ...> ]
      ...> Application.put_env(:flux_redis, :hash_storages, hash_storages)
      ...> animals = [
      ...>   %{family: "Canidae", phylum: "Chordata"},
      ...>   %{family: "Felidae", phylum: "Chordata"}
      ...> ]
      ...> car = %{brand: "Honda", model: "Civic", year: 2019}
      ...> users = [%{id: 1, name: "John"}, %{id: 2, name: "Doe"}]
      ...> pipeline_payload = [
      ...>   FluxRedis.HashManager.Commands.set_many!(:animals, animals)
      ...>   FluxRedis.HashManager.Commands.set!(:cars, car)
      ...>   FluxRedis.HashManager.Commands.set_many!(:users, users)
      ...>   FluxRedis.HashManager.Commands.delete(:users, 1)
      ...>   FluxRedis.HashManager.Commands.count(:animals)
      ...> ]
      ...> FluxRedis.PipelineManager.pipeline!(pipeline_payload)
      [2, :set, 2, :ok, 2]
  """
  @doc since: "0.0.2"
  @spec pipeline!(list({command, parser}), keyword) :: :ok | list(any)
  def pipeline!(pipeline_payload, opts \\ []) do
    pipeline_payload
    |> Enum.unzip()
    |> Executer.execute_pipeline!(opts)
  end
end
